/* Copyright Siemens, 2023 */


#include <gtest/gtest.h>
#include <vector>

#include "../../implementation/message/include/deserializer.hpp"
#include "../../implementation/message/include/message_impl.hpp"
#include "../../implementation/message/include/serializer.hpp"
#include <vsomeip/vsomeip.hpp>

using namespace vsomeip;

TEST(message_test, someip) {
  {
    std::vector<vsomeip::byte_t> expectedData{
        0x13, 0x88, 0x01, 0x90, 0x00, 0x00, 0x00, 0x0C, 0x00, 0x01,
        0x00, 0x01, 0x01, 0x00, 0x02, 0x00, 0x00, 0x01, 0x00, 0x09};

    auto message = runtime::get()->create_notification(true, false);
    message->set_service(5000);
    message->set_instance(600);
    message->set_method(400);
    message->set_client(1);
    message->set_session(1);
    std::shared_ptr<payload> its_payload =
        vsomeip::runtime::get()->create_payload();
    std::vector<vsomeip::byte_t> its_payload_data{0x00, 0x01, 0x00, 0x09};
    its_payload->set_data(its_payload_data);
    message->set_payload(its_payload);
    serializer its_serializer(5);
    its_serializer.serialize(message.get());
    auto serializedData = its_serializer.get_data();

    EXPECT_EQ(expectedData.size(), its_serializer.get_size());
    EXPECT_FALSE(
        std::memcmp(expectedData.data(), serializedData, expectedData.size()));
  }

  {
    std::vector<vsomeip::byte_t> serializedData{
        0x13, 0x88, 0x01, 0x90, 0x00, 0x00, 0x00, 0x0C, 0x00, 0x01,
        0x00, 0x01, 0x01, 0x00, 0x02, 0x00, 0x00, 0x01, 0x00, 0x09};
    std::vector<vsomeip::byte_t> expectedPayloadData{0x00, 0x01, 0x00, 0x09};

    deserializer its_deserializer(5);
    its_deserializer.set_data(serializedData.data(), serializedData.size());
    std::shared_ptr<message_impl> message(
        its_deserializer.deserialize_message(true));
    std::shared_ptr<payload> its_payload = message->get_payload();
    auto payloadData = its_payload->get_data();

    EXPECT_EQ(message->get_length(), 12);
    EXPECT_EQ(message->get_service(), 5000);
    EXPECT_EQ(message->get_method(), 400);
    EXPECT_EQ(message->get_message_type(), message_type_e::MT_NOTIFICATION);
    EXPECT_TRUE(message->is_someip_message());
    EXPECT_EQ(expectedPayloadData.size(), its_payload->get_length());
    EXPECT_FALSE(std::memcmp(expectedPayloadData.data(), payloadData,
                             expectedPayloadData.size()));
  }
}

TEST(message_test, signalbased) {
  {
    std::vector<vsomeip::byte_t> expectedData{
        0x13, 0x88, 0x01, 0x90, 0x00, 0x00, 0x00, 0x04, 0x00, 0x01, 0x00, 0x09};

    auto message = runtime::get()->create_notification(false, false);
    message->set_service(5000);
    message->set_method(400);
    std::shared_ptr<payload> its_payload =
        vsomeip::runtime::get()->create_payload();
    std::vector<vsomeip::byte_t> its_payload_data{0x00, 0x01, 0x00, 0x09};
    its_payload->set_data(its_payload_data);
    message->set_payload(its_payload);
    serializer its_serializer(5);
    its_serializer.serialize(message.get());
    auto serializedData = its_serializer.get_data();

    EXPECT_EQ(expectedData.size(), its_serializer.get_size());
    EXPECT_FALSE(
        std::memcmp(expectedData.data(), serializedData, expectedData.size()));
  }

  {
    std::vector<vsomeip::byte_t> serializedData{
        0x13, 0x88, 0x01, 0x90, 0x00, 0x00, 0x00, 0x04, 0x00, 0x01, 0x00, 0x09};
    std::vector<vsomeip::byte_t> expectedPayloadData{0x00, 0x01, 0x00, 0x09};

    deserializer its_deserializer(5);
    its_deserializer.set_data(serializedData.data(), serializedData.size());
    std::shared_ptr<message_impl> message(
        its_deserializer.deserialize_message(false));
    std::shared_ptr<payload> its_payload = message->get_payload();
    auto payloadData = its_payload->get_data();

    EXPECT_EQ(message->get_length(), 4);
    EXPECT_EQ(message->get_service(), 5000);
    EXPECT_EQ(message->get_method(), 400);
    EXPECT_FALSE(message->is_someip_message());
    EXPECT_EQ(message->get_message_type(), message_type_e::MT_NOTIFICATION);
    EXPECT_EQ(expectedPayloadData.size(), its_payload->get_length());
    EXPECT_FALSE(std::memcmp(expectedPayloadData.data(), payloadData,
                             expectedPayloadData.size()));
  }
}

int main(int argc, char **argv) {
  ::testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
