/* Copyright Siemens, 2023 */

#include <chrono>
#include <condition_variable>
#include <iostream>
#include <sstream>
#include <thread>
#include <future>

#include <gtest/gtest.h>

#include <vsomeip/vsomeip.hpp>
#include <vsomeip/internal/logger.hpp>
#include "signal_based_service_info.hpp"

using namespace vsomeip;

vsomeip::reliability_type_e reliability_type = vsomeip::reliability_type_e::RT_UNKNOWN;
class signal_based_test_server {
public:
    signal_based_test_server(struct signal_based_test::service_info _service_info) :
            service_info_(_service_info),
            app_(vsomeip::runtime::get()->create_application("signal_based_server")),
            offer_thread_(std::bind(&signal_based_test_server::run, this))
        {
        if (!app_->init()) {
            ADD_FAILURE() << "[" << app_->get_name() << "] Couldn't initialize application";
            return;
        }
        // register a handler for the application state
        app_->register_state_handler(
                std::bind(&signal_based_test_server::on_state, this,
                        std::placeholders::_1));
        // register a handler for the request
        app_->register_message_handler(service_info_.service_id,
                service_info_.instance_id, service_info_.method_id,
                std::bind(&signal_based_test_server::on_request, this,
                        std::placeholders::_1));
        // register a handler for the shutdown message
        app_->register_message_handler(service_info_.service_id,
                service_info_.instance_id, service_info_.shutdown_method_id,
                std::bind(&signal_based_test_server::on_shutdown_method_called, this,
                        std::placeholders::_1));
        std::promise<bool> its_promise;
        application_thread_ = std::thread([&](){
            its_promise.set_value(true);
            app_->start();
        });
        EXPECT_TRUE(its_promise.get_future().get());
    }

    ~signal_based_test_server() {
        
        offer_thread_.join();
        application_thread_.join();
    }

    void on_state(vsomeip::state_type_e _state) {
        VSOMEIP_INFO << "[" << app_->get_name() << "] Application is "
        << (_state == vsomeip::state_type_e::ST_REGISTERED ?
                "registered." : "deregistered.");

        if (_state == vsomeip::state_type_e::ST_REGISTERED) {
            std::lock_guard<std::mutex> its_lock(mutex_);
            condition_.notify_one();
        }
    }

    void on_request(const std::shared_ptr<vsomeip::message> &_message) {
        static_cast<void>(_message);
        VSOMEIP_INFO << "[" << app_->get_name() << "] Received a request from the signal based client application to send the data";
        std::this_thread::sleep_for(std::chrono::milliseconds(500));

        std::vector<byte_t> data{1,2,3,4,5,6,7,8,9,10,44,44,33,11,5,5,7,67,67,67,6};
        std::shared_ptr<vsomeip::payload> its_payload = vsomeip::runtime::get()->create_payload();
        its_payload->set_data(data);
        app_->notify(service_info_.service_id, service_info_.instance_id,
                service_info_.event_id, its_payload);
        std::vector<byte_t> data2{1,2};
        std::shared_ptr<vsomeip::payload> its_payload2 = vsomeip::runtime::get()->create_payload();
        its_payload2->set_data(data2);
        app_->notify(service_info_.service_id, service_info_.instance_id,
                service_info_.event_id, its_payload2);
        std::vector<byte_t> data3{};
        std::shared_ptr<vsomeip::payload> its_payload3 = vsomeip::runtime::get()->create_payload();
        its_payload3->set_data(data3);
        app_->notify(service_info_.service_id, service_info_.instance_id,
                service_info_.event_id, its_payload3);
    }

    void offer() {
        app_->offer_service(service_info_.service_id,
                service_info_.instance_id);
        std::set<vsomeip::eventgroup_t> its_eventgroups;
        its_eventgroups.insert(service_info_.eventgroup_id);
        // offer an event with signal based serialization flag enabled
        app_->offer_event(service_info_.service_id, service_info_.instance_id,
                service_info_.event_id, its_eventgroups, false,
                vsomeip::event_type_e::ET_EVENT, std::chrono::milliseconds::zero(),
                false, true, nullptr, reliability_type);
        app_->register_subscription_handler(service_info_.service_id,
                service_info_.instance_id, service_info_.eventgroup_id,
                std::bind(&signal_based_test_server::subscription_handler, this,
                        std::placeholders::_1, std::placeholders::_2,
                        std::placeholders::_3, std::placeholders::_4));
    }

    void run() {
        VSOMEIP_DEBUG << "[" << app_->get_name() << "] application is Running";
        std::unique_lock<std::mutex> its_lock(mutex_);
        condition_.wait(its_lock);
        VSOMEIP_DEBUG << "[" << app_->get_name() << "] is Offering the service and the event";
        offer();
    }

    void stop() {
        VSOMEIP_INFO << "[" << app_->get_name() << "] : " << "Shutdown";
        app_->stop_offer_service(service_info_.service_id, service_info_.instance_id);
        app_->clear_all_handler();
        app_->stop();
    }

    void on_shutdown_method_called(const std::shared_ptr<vsomeip::message> &_message) {
        VSOMEIP_INFO << "[" << app_->get_name() << "]" << " Received a shutdown request from signal based client application.";
        (void)_message;
        stop();
    }

    bool subscription_handler(vsomeip::client_t _client, std::uint32_t _uid, std::uint32_t _gid, bool _subscribed) {
        (void)_uid;
        (void)_gid;
        VSOMEIP_INFO << "[" << app_->get_name() << "]" << " client: [signal_based_client] with id 0x" << std::hex << _client
                << ((_subscribed) ? " subscribed." : "unsubscribed.");
        return true;
    }

private:
    struct signal_based_test::service_info service_info_;
    std::shared_ptr<vsomeip::application> app_;
    std::mutex mutex_;
    std::condition_variable condition_;
    std::thread offer_thread_;
    std::thread application_thread_;
};

#ifndef _WIN32
int main(int argc, char** argv)
{
    if (argc > 0) {
        if (std::string("TCP")== std::string(argv[0])) {
            VSOMEIP_INFO << "TCP is used";
            reliability_type = vsomeip::reliability_type_e::RT_RELIABLE;
        } else{
            reliability_type = vsomeip::reliability_type_e::RT_UNRELIABLE;
        }
    }
    else{
        reliability_type = vsomeip::reliability_type_e::RT_RELIABLE;
    }
    signal_based_test_server its_server(signal_based_test::service);  
}
#endif




