/* Copyright Siemens, 2023 */

#include <chrono>
#include <condition_variable>
#include <iostream>
#include <sstream>
#include <thread>
#include <future>

#include <gtest/gtest.h>
#include <cstring>
#include <vsomeip/vsomeip.hpp>
#include <vsomeip/internal/logger.hpp>
#include "signal_based_service_info.hpp"

using namespace vsomeip;

vsomeip::reliability_type_e reliability_type = vsomeip::reliability_type_e::RT_UNKNOWN;
class signal_based_test_client {
public:
    signal_based_test_client(struct signal_based_test::service_info _service_info) :
            service_info_(_service_info),
            app_(vsomeip::runtime::get()->create_application("signal_based_client")),
            received_responses_(0),
            stop_thread_(std::bind(&signal_based_test_client::stop, this)),
            send_thread_(std::bind(&signal_based_test_client::run, this))
        {
        if (!app_->init()) {
            ADD_FAILURE() << "[" << app_->get_name() << "] Couldn't initialize application";
            return;
        }
        app_->register_state_handler(
                std::bind(&signal_based_test_client::on_state, this,
                        std::placeholders::_1));

        std::promise<bool> its_promise;
        application_thread_ = std::thread([&](){
            its_promise.set_value(true);
            app_->start();
        });
        EXPECT_TRUE(its_promise.get_future().get());
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
    }

    ~signal_based_test_client() {
        stop_thread_.join();
        send_thread_.join();
        application_thread_.join();
    }

    void on_state(vsomeip::state_type_e _state) {
        VSOMEIP_INFO << "[" << app_->get_name() << "] Application is "
        << (_state == vsomeip::state_type_e::ST_REGISTERED ?
                "registered." : "deregistered.");

        if (_state == vsomeip::state_type_e::ST_REGISTERED) {
            std::lock_guard<std::mutex> its_lock(mutex_);
            condition_.notify_one();
        }
    }

    void request() {
        // register availability for all other services and request their event.
        std::set<vsomeip::eventgroup_t> its_eventgroups;
        its_eventgroups.insert(service_info_.eventgroup_id);
        app_->request_event(service_info_.service_id,
                service_info_.instance_id,
                service_info_.event_id, its_eventgroups, false,
                vsomeip::event_type_e::ET_EVENT,
                reliability_type);
        app_->register_message_handler(service_info_.service_id,
                service_info_.instance_id,
                service_info_.event_id,
                std::bind(&signal_based_test_client::on_message, this,
                        std::placeholders::_1));
        app_->register_availability_handler(service_info_.service_id,
                service_info_.instance_id,
                std::bind(&signal_based_test_client::on_availability, this,
                        std::placeholders::_1, std::placeholders::_2,
                        std::placeholders::_3));
        app_->request_service(service_info_.service_id,
                service_info_.instance_id);
    }

    void run() {
        VSOMEIP_DEBUG << "[" << app_->get_name() << "] application is Running";
        std::unique_lock<std::mutex> its_lock(mutex_);
        condition_.wait(its_lock);
        VSOMEIP_DEBUG << "[" << app_->get_name() << "] is requesting the service and the event";
        request();
    }

    void on_availability(vsomeip::service_t _service,
                         vsomeip::instance_t _instance, bool _is_available) {
        VSOMEIP_INFO << "[" << app_->get_name() << "] Service [" << std::setw(4)
        << std::setfill('0') << std::hex << _service << "." << _instance
        << "] is " << (_is_available ? "available":"not available") << ".";
        std::lock_guard<std::mutex> its_lock(mutex_);
        if(_is_available) {
            std::this_thread::sleep_for(std::chrono::milliseconds(1000));
            app_->subscribe(service_info_.service_id,
                    service_info_.instance_id,
                    service_info_.eventgroup_id, 0x0,
                    service_info_.event_id);
            std::shared_ptr<vsomeip::message> its_req = vsomeip::runtime::get()->create_request();
            its_req->set_service(service_info_.service_id);
            its_req->set_instance(service_info_.instance_id);
            its_req->set_method(service_info_.method_id);
            app_->send(its_req);
        } else {
        }
    }

    void stop() {
        std::unique_lock<std::mutex> its_lock(stop_mutex_);
        stop_condition_.wait(its_lock);
        VSOMEIP_INFO << "[" << app_->get_name() << "] : " << "Shutdown..";
        app_->clear_all_handler();
        app_->stop();
    }

    void on_message(const std::shared_ptr<vsomeip::message> &_message) {
        std::lock_guard<std::mutex> its_lock(mutex_);
        std::vector<byte_t> expectedPayloadData{1,2,3,4,5,6,7,8,9,10,44,44,33,11,5,5,7,67,67,67,6};
        std::vector<byte_t> expectedPayloadData2{1,2};
        ++received_responses_;
        EXPECT_EQ(service_info_.service_id, _message->get_service());
        EXPECT_EQ(service_info_.event_id, _message->get_method());
        EXPECT_EQ(service_info_.instance_id, _message->get_instance());
        EXPECT_FALSE(_message->is_someip_message());
        std::shared_ptr<payload> its_payload = _message->get_payload();

        VSOMEIP_INFO << "[" << app_->get_name() << "] Received a message from [signal_based_server] application and number of received messages : " << received_responses_;
        if(received_responses_ == 1)
        {
            EXPECT_EQ(21, its_payload->get_length());
            auto payloadData = its_payload->get_data();
            EXPECT_FALSE(std::memcmp(expectedPayloadData.data(), payloadData, expectedPayloadData.size()));
        }
        else if(received_responses_ == 2)
        {
            EXPECT_EQ(2, its_payload->get_length());
            auto payloadData = its_payload->get_data();
            EXPECT_FALSE(std::memcmp(expectedPayloadData2.data(), payloadData, expectedPayloadData2.size()));
        }
        else if(received_responses_ == 3)
        {
            EXPECT_EQ(0, its_payload->get_length());
            VSOMEIP_INFO << "[" << app_->get_name() << "] Send Shutdown request to the [signal_based_server] application";
            std::shared_ptr<vsomeip::message> its_req = vsomeip::runtime::get()->create_request();
            its_req->set_service(service_info_.service_id);
            its_req->set_instance(service_info_.instance_id);
            its_req->set_method(service_info_.shutdown_method_id);
            app_->send(its_req);
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
            stop_condition_.notify_one();
        }
    }

private:
    struct signal_based_test::service_info service_info_;
    std::shared_ptr<vsomeip::application> app_;
    std::mutex mutex_;
    std::condition_variable condition_;
    std::mutex stop_mutex_;
    std::condition_variable stop_condition_;
    std::atomic<std::uint32_t> received_responses_;
    std::thread stop_thread_;
    std::thread send_thread_;
    std::thread application_thread_;
};

TEST(signal_based_test, notify_and_notify_one) {
    signal_based_test_client its_client(signal_based_test::service);
}



#ifndef _WIN32
int main(int argc, char** argv)
{

    if (argc > 0) {
        if (std::string("TCP")== std::string(argv[0])) {
            VSOMEIP_INFO << "TCP is used";
            reliability_type = vsomeip::reliability_type_e::RT_RELIABLE;
        } else{
            reliability_type = vsomeip::reliability_type_e::RT_UNRELIABLE;
        }
    }
    else{
        reliability_type = vsomeip::reliability_type_e::RT_RELIABLE;
    }
    ::testing::InitGoogleTest(&argc, argv);
    return RUN_ALL_TESTS();
}
#endif




