#!/bin/sh
# Copyright (C) 2015-2017 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Copyright Siemens, 2023

# Purpose: This script is needed to start the client and service with
# one command. This is necessary as ctest - which is used to run the
# tests - isn't able to start two binaries for one testcase. Therefore
# the testcase simply executes this script. This script then runs client
# and service and checks that both exit sucessfully.

FAIL=0

# Parameter 1: the pid to check
# Parameter 2: number of TCP/UDP sockets the process should have open
check_tcp_udp_sockets_are_open ()
{
    # Check that the passed pid/process does listen on at least one TCP/UDP socket 
    # awk is used to avoid the case when a inode number is the same as a PID. The awk
    # program filters the sockstat output down to the protocol (5th field) and 
    # the PID/Program name (3rd field) fields.
    SERVICE_SOCKETS_LISTENING=$(sockstat -46 2> /dev/null | awk '{print $3 "\t"  $5}' | grep $1 | wc -l)
    if [ $SERVICE_SOCKETS_LISTENING -lt $2 ]
    then
        ((FAIL+=1))
    fi
}

# Parameter 1: the pid to check
check_tcp_udp_sockets_are_closed ()
{
    # Check that the passed pid/process does not listen on any TCP/UDP socket 
    # or has any active connection via a TCP/UDP socket
    # awk is used to avoid the case when a inode number is the same as a PID. The awk
    # program filters the sockstat output down to the protocol (5th field) and 
    # the PID/Program name (3rd field) fields.
    SERVICE_SOCKETS_LISTENING=$(sockstat -46 2> /dev/null | awk '{print $3 "\t"  $5}' | grep $1 | wc -l)
    if [ $SERVICE_SOCKETS_LISTENING -ne 0 ]
    then
        ((FAIL+=1))
    fi

    SERVICE_SOCKETS_LISTENING=$(sockstat -46 2> /dev/null | awk '{print $3 "\t"  $5}' | grep $1 | wc -l)
    if [ $SERVICE_SOCKETS_CONNECTED -ne 0 ]
    then
        ((FAIL+=1))
    fi
}

# Start the service
export VSOMEIP_APPLICATION_NAME=external_local_payload_test_service
export VSOMEIP_CONFIGURATION=external_local_payload_test_service.json
./payload_test_service &
SERIVCE_PID=$!
sleep 3

# The service should listen on a TCP and UDP socket now
check_tcp_udp_sockets_are_open $SERIVCE_PID 2

# Start the client which sends messages over local UDS
export VSOMEIP_APPLICATION_NAME=external_local_payload_test_client_local
export VSOMEIP_CONFIGURATION=external_local_payload_test_client_local.json
./payload_test_client --dont-shutdown-service &
CLIENT_PID=$!
sleep 3

check_tcp_udp_sockets_are_open $SERIVCE_PID 2
check_tcp_udp_sockets_are_closed $CLIENT_PID

# Wait until client is finished
wait $CLIENT_PID
exit_status=$?
if [ $exit_status != 0 ] && [ $exit_status != 127 ]; then
    echo "Fail status $exit_status"
    FAIL=$FAIL+1
fi

# Display a message to show the user that he must now call the external client
# to finish the test successfully
ssh -o "StrictHostKeyChecking=no" root@172.31.1.2 "cd /data/home/qnxuser/build/test && ./external_local_payload_test_client_external_start_qnx.sh" &

# The service should still listen on a TCP and UDP socket
sleep 3
check_tcp_udp_sockets_are_open $SERIVCE_PID 2

# Wait until service is finished
# The client remotely shuts down the service if he has successfully transmitted
# all the packets with different payloads. Therefore we can assume that everything
# went well, even if we can only check the exit code of the service here.

# Fail gets incremented if either client or service exit
# with a non-zero exit code
wait $SERIVCE_PID
exit_status=$?
if [ $exit_status != 0 ] && [ $exit_status != 127 ]; then
    echo "Fail status $exit_status"
    FAIL=$FAIL+1
fi


# Start the service for payload test with tcp
export VSOMEIP_APPLICATION_NAME=external_local_payload_test_service
export VSOMEIP_CONFIGURATION=external_local_payload_test_service.json
./payload_test_service --tcp &
SERIVCE_PID=$!

# The service should listen on a TCP and UDP socket now
sleep 3
check_tcp_udp_sockets_are_open $SERIVCE_PID 2


# Wait until service is finished
# The client remotely shuts down the service if he has successfully transmitted
# all the packets with different payloads. Therefore we can assume that everything
# went well, even if we can only check the exit code of the service here.

# Fail gets incremented if either client or service exit
# with a non-zero exit code
wait $SERIVCE_PID
exit_status=$?
if [ $exit_status != 0 ] && [ $exit_status != 127 ]; then
    echo "Fail status $exit_status"
    FAIL=$FAIL+1
fi

# Check if client and server both exited sucessfully and the service didnt't
# have any open TCP/UDP sockets
if [ $FAIL -eq 0 ]
then
    exit 0
else
    exit 1
fi
